==> mdcached-1.0.2.tgz: mdcached-1.0.2/mdcached/DataStore.h <==
/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2010. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: DataStore.h 410 2012-01-20 15:21:53Z ivoras $
 */

#ifndef _DATASTORE_H_
#define _DATASTORE_H_

#include <exception>
#include <set>
#include <vector>
#include <algorithm>
#include <utility>

#include <string.h>

#include "mc_protocol.h"
#include "LockPool.h"

#ifdef HASH_DJB2
#include "hash.h"
#else
#include "MurmurHash3.h"
#define hash32_buf(buf, size) (MurmurHash3_32(buf, size, 0xA00AAFFA))
#endif

/** Is integer a power of 2? */
#define PO2(x) ((x&(x-1))==0)

#ifndef DEFAULT_N_DATA_NODES
#define DEFAULT_N_DATA_NODES 8
#endif

#ifndef N_ITEM_BUCKETS
#define N_ITEM_BUCKETS 256
#endif

#define ITEM_BUCKETS_MASK (N_ITEM_BUCKETS-1)
#define ITEM_BUCKET_INDEX(name,name_size) (hash32_buf(name, name_size) & ITEM_BUCKETS_MASK)

#ifndef N_TAG_BUCKETS
#define N_TAG_BUCKETS 64
#endif

#define TAG_BUCKETS_MASK (N_TAG_BUCKETS-1)
#define TAG_BUCKET_INDEX(nt) (nt & TAG_BUCKETS_MASK)


/* The prewarm cache dump constants and structures */
#define PREWARM_SIGNATURE	0xab666ffe
#define PREWARM_FORMAT		0x01
#define PREWARM_FLAG_COMPRESSED	0x01
#define PREWARM_FLAG_SANE	0x02
struct mdcached_prewarm_file_header {
	uint32_t 		signature;
	uint16_t		format_version;
	uint16_t		flags;
	uint16_t		n_item_buckets;
	uint16_t		n_tag_buckets;
	uint32_t		n_records;
};


enum LOCK_TYPE { LT_BOGUS, LT_ITEMS, LT_TAGS };

struct dh_items_bucket;
struct dh_item_tree {
	struct dh_data *rbh_root;
};
struct tag_value_tree {
	struct tag_value *rbh_root;
};

struct dh_data {
	RB_ENTRY(dh_data)	 linkage;
	/* All the following pointers actually point somewhere inside the
	 * inpkt blob of data and are used as cached values. */
	char 			*name;
	unsigned short int	name_size;
	unsigned short int 	n_tags;
	char			*data;
	unsigned int		data_size;
	struct mc_tag		*tags;
	time_t			mtime, exptime;
	struct dh_items_bucket	*bucket;	/* backpoint to the hash bucket */
	/* end cached data fields */
	struct mc_data_entry	*inpkt;
#ifdef DEEP_STATS
	/* For dh_data structures, DEEP_STATS covers both read and write stats */
	unsigned int		stat_n_write_ops;
	unsigned int		stat_n_read_ops;
#endif
};


/**
 * Each item hash bucket contains the head of a tree containing dh_data nodes,
 * protected by a rwlock.
 */
struct dh_items_bucket {
	unsigned int		index;
	struct dh_item_tree	tree_head;
	unsigned int		memory_used;
	unsigned int		stat_n_write_ops;
#ifdef DEEP_STATS
	unsigned int		stat_n_read_ops;
#endif
};


/**
 * To record a link between a tag value and an struct dh_data item, we also need
 * to record its bucket in the same data structure so it is covered by the same
 * lock (referencing the bucket as item->bucket is racey).
 */
struct dh_item_value_entry {
	struct dh_data			*item;
	struct dh_items_bucket		*bucket;
};

/*
 * HOW THE TAGS ARE STORED
 * -----------------------
 *
 * The tag buckets are indexed by the tag key. Tag buckets realize the granularity
 * of locks.
 *
 * Each tag bucket has a flat array of struct tag_key keys, which holds the
 * actual keys. It is expected that each tag bucket (i.e. key hash collisions) will
 * contain a small number of invididual keys (on the order of 8) for efficiency,
 * i.e. that the keypace is small.
 *
 * Each of the entries in keys contains a root of the tag_value_tree named
 * tree_head. This tree contains the values part of the tags.
 *
 * Each entry in the tag_value_tree contains a flat vitems array (which has
 * items_size elements) containing pointers to the actual data items (or NULL).
 */


/**
 * A "tag value" structure, with pointers to items containing this specific
 * key-value pair.
 */
struct tag_value {
	RB_ENTRY(tag_value)		linkage;
	signed int			value;
	unsigned int			items_size;
	struct dh_item_value_entry	*vitems;
};


static inline int
sh_tag_value_cmp(struct tag_value *a, struct tag_value *b)
{
	if (a->value == b->value)
		return 0;
	else
		return a->value < b->value ? -1 : 1;
}


/**
 * The root of a tree containing tag values for this specific key. Key must
 * be non-zero.
 */
struct tag_key {
	tag_key_t			key;
	struct tag_value_tree		tree_head;
};

RB_PROTOTYPE(tag_value_tree, tag_value, linkage, sh_tag_value_cmp);


/**
 * Each tag bucket contains an array of tag_key structures, each of which
 * contains a tree of tag_value structures.
 */
struct dh_tags_bucket {
	unsigned int			index;
	unsigned int			keys_size;
	struct tag_key			*keys;
	unsigned int			stat_n_write_ops;
#ifdef DEEP_STATS
	unsigned int			stat_n_read_ops;
#endif
};


/*
 * Locking notes:
 * Locking in the direction: item bucket lock held -> single tag locks acquired
 * and released can go in blocking mode without deadlocks. All other directions
 * and forms of locking, including acquirings of multiple locks of the same
 * level, must go through the retry (trylock) protocol to avoid deadlocks.
 */


static inline int
sh_item_cmp(struct dh_data *a, struct dh_data *b)
{
	int res;
/*
	assert(a != NULL);
	assert(b != NULL);
*/
	if (a->name_size != b->name_size)
		res = (int)(a->name_size) - (int)(b->name_size);
	else
		res = memcmp(a->name, b->name, a->name_size);
	return res;
}

struct dh_item_tree;

RB_PROTOTYPE(dh_item_tree, dh_data, linkage, __unused);
RB_PROTOTYPE(tag_value_tree, tag_value, linkage, __unused);


class MDCacheException : public std::exception {
private:
	char *msg;
public:
	MDCacheException(char *msg) : exception() {
		this->msg = msg;
	}
};


/**
 * DataStore is the principal data storage class.
 *
 * The locking dance described:
 *
 *	items_locks and tags_locks are the lock pools for items and tags; locks
 *	are for all intents and purposes identified with their numeric IDs, which
 *	are also indexes to items and tags arrays they protect.
 *
 *	Since they are abstracted as integers, they are passed around as such to
 *	other backends in the mesh (if any) by the BackendMesh via the lock
 *	notifiers.
 *
 *	The assumption is (XXX: is it still valid?) that all single lock operations
 *	must succeed (e.g. single item / tag lock operations) since they are blocking,
 *	but all multiple-lock operations must implement the trylock-and-rollback protocol
 *	to avoid deadlocks.
 *
 *	This will probably need to be revised when dealing with partition tolerance.
 */
class DataStore {
private:
	struct dh_items_bucket	items[N_ITEM_BUCKETS];
	LockPool<N_ITEM_BUCKETS> items_locks;
	struct dh_tags_bucket	tags[N_TAG_BUCKETS];
	LockPool<N_TAG_BUCKETS>	tags_locks;
	
	unsigned stat_reap_dead_count;
	unsigned stat_reap_memory_count;

	void remove_item_and_free(struct dh_data *d);
	void data_cleanup(struct dh_data *d);
	void tags_cleanup(struct dh_data *d);
	bool tag_link(struct dh_data *d, struct mc_tag *t);
	bool tag_unlink(struct dh_data *d, struct mc_tag *t);
	int get_items_buckets_by_tag_key_values(struct tag_key *tk, const tag_value_t *tvalues, const unsigned values_count,
		unsigned **buckets, unsigned *n_buckets);
	bool items_buckets_rdlock(unsigned *buckets, unsigned n_buckets);
	bool items_buckets_wrlock(unsigned *buckets, unsigned n_buckets);
	void items_buckets_unlock(unsigned *buckets, unsigned n_buckets);
	bool tags_buckets_wrlock(unsigned *buckets, unsigned n_buckets);
	void tags_buckets_unlock(unsigned *buckets, unsigned n_buckets);
	bool data_item_safe_lock_op(struct dh_data *d, enum LOCK_OP lop, bool doitem, bool dotags);
	bool multidata_append(struct mc_resp_multidata **md, struct mc_data_entry *item);
	bool multistring_append(struct mc_resp_multistring **ms, const char *str, const unsigned short int str_size);
	int put_nolock(struct dh_items_bucket *ib, struct dh_data **d);
	struct mc_data_entry* get_data_nolock(char *name, unsigned int name_size);
	int del_data_name_nolock(const char *name, const unsigned int name_size);
	int del_nolock(struct dh_items_bucket *ib, struct dh_data *d);
	int prewarm_cache_dump_bucket(FILE *f, unsigned bi);
	bool prewarm_cache_read_put_record(FILE *f, bool do_lock);
	struct dh_data *internal_add_record_nolock(const char *name, unsigned name_size, const char *data, unsigned data_size, struct mc_tag *tags, unsigned n_tags);
	bool item_is_alive(struct dh_data *d);
public:
	unsigned int 		n_item_buckets;
	unsigned int		n_tag_buckets;

	static void fill_dh_data(struct dh_data *entry, struct mc_data_entry *inpkt);

	DataStore() throw(MDCacheException);
	~DataStore();

	void set_lock_notifiers(LockNotifier *ilock_notifier, LockNotifier *tlock_notifier);
	void unset_lock_notifiers();

	enum LOCK_RESULT _internalLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE);

	bool prewarm_cache_dump(char *filename, bool atomic);
	bool prewarm_cache_read(char *filename, bool atomic);

	/** Put a cooked dh_data structure into the cache */
	int put(struct dh_data *d);

	/** Get a cooked dh_data structure by name */
	struct dh_data* get(char *name, unsigned int name_len);

	/** Get a mc_data_entry structure by name */
	struct mc_data_entry* get_data(char *name, unsigned int name_size);

	/** Get a mc_data_entry structure by name but leave the tree locked for reading (to support zero-copy) */
	struct mc_data_entry* get_data_and_lock(char *name, unsigned int name_size, unsigned int *lock_cookie);

	/** Unlock the tree locked by dh_get_data_and_lock */
	void got_data_unlock(unsigned int lock_cookie);

	/** Retrieve multiple data items matching given key and any of the given values */
	struct mc_resp_multidata* get_multidata_by_tag_values(const tag_key_t tkey, tag_value_t *values, const unsigned int values_count);

	/**
	 * Retrieve all the record keys (key names) stored in the cache. Depending
	 * on the flags this may happen with all buckets locked or one by one.
	 */
	struct mc_resp_multistring *get_all_keys(const unsigned flags);

	/** Delete a data item identified by name */
	int del_data_name(const char *name, const unsigned int name_size);

	/** Delete data items identified by key and a list of values */
	int del_data_tag_values(const tag_key_t tag_key, const unsigned int num_values, const tag_value_t *values);

	/** One of the simple atomic ops. */
	struct mc_resp_atomic* atomic_op(const int op, const char *name, const unsigned int name_size, const int64_t arg1, const int64_t arg2);

	/** Multiple-put: puts multiple records into the cache */
	int mput(struct dh_data **entries, unsigned n_records, unsigned flags);

	/** Multiple-get: gets multiple records from the cache by record keys */
	struct mc_resp_multidata* mget(struct mc_key_name **keys, unsigned n_names, unsigned flags);

	/** Multi-del: deletes multile records from the cache by record keys */
	int mdel(struct mc_key_name **keys, unsigned n_names, unsigned flags);
	
	/** Tstack pop: returnes & removes a record from a virtual stack specified by the tag key & value */
	struct mc_data_entry* tstack_pop(tag_key_t tkey, tag_value_t tval);
	
	/** Perform book-keeping; returns the number of records reaped */
	unsigned reap_dead_items();
	unsigned reap_free_memory(uint64_t memory_limit);
};

#endif

==> mdcached-1.0.3.tgz: mdcached-1.0.3/mdcached/DataStore.h <==
/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2010. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: DataStore.h 437 2012-03-05 22:08:46Z ivoras $
 */

#ifndef _DATASTORE_H_
#define _DATASTORE_H_

#include <exception>
#include <set>
#include <vector>
#include <algorithm>
#include <utility>

#include <string.h>

#include "mc_protocol.h"
#include "LockPool.h"

#ifdef HASH_DJB2
#include "hash.h"
#else
#include "MurmurHash3.h"
#define hash32_buf(buf, size) (MurmurHash3_32(buf, size, 0xA00AAFFA))
#endif

#ifndef DEFAULT_N_DATA_NODES
#define DEFAULT_N_DATA_NODES 8
#endif

#ifndef N_ITEM_BUCKETS
#define N_ITEM_BUCKETS 256
#endif

#define ITEM_BUCKETS_MASK (N_ITEM_BUCKETS-1)
#define ITEM_BUCKET_INDEX(name,name_size) (hash32_buf(name, name_size) & ITEM_BUCKETS_MASK)

#ifndef N_TAG_BUCKETS
#define N_TAG_BUCKETS 64
#endif

#define TAG_BUCKETS_MASK (N_TAG_BUCKETS-1)
#define TAG_BUCKET_INDEX(nt) (nt & TAG_BUCKETS_MASK)


/* The prewarm cache dump constants and structures */
#define PREWARM_SIGNATURE	0xab666ffe
#define PREWARM_FORMAT		0x01
#define PREWARM_FLAG_COMPRESSED	0x01
#define PREWARM_FLAG_SANE	0x02
struct mdcached_prewarm_file_header {
	uint32_t 		signature;
	uint16_t		format_version;
	uint16_t		flags;
	uint16_t		n_item_buckets;
	uint16_t		n_tag_buckets;
	uint32_t		n_records;
};


enum LOCK_TYPE { LT_BOGUS, LT_ITEMS, LT_TAGS };

struct dh_items_bucket;
struct dh_item_tree {
	struct dh_data *rbh_root;
};
struct tag_value_tree {
	struct tag_value *rbh_root;
};

struct dh_data {
	RB_ENTRY(dh_data)	 linkage;
	/* All the following pointers actually point somewhere inside the
	 * inpkt blob of data and are used as cached values. */
	char 			*name;
	unsigned short int	name_size;
	unsigned short int 	n_tags;
	char			*data;
	unsigned int		data_size;
	struct mc_tag		*tags;
	time_t			mtime, exptime;
	struct dh_items_bucket	*bucket;	/* backpoint to the hash bucket */
	/* end cached data fields */
	struct mc_data_entry	*inpkt;
#ifdef DEEP_STATS
	/* For dh_data structures, DEEP_STATS covers both read and write stats */
	unsigned int		stat_n_write_ops;
	unsigned int		stat_n_read_ops;
#endif
};


/**
 * Each item hash bucket contains the head of a tree containing dh_data nodes,
 * protected by a rwlock.
 */
struct dh_items_bucket {
	unsigned int		index;
	struct dh_item_tree	tree_head;
	unsigned int		memory_used;
	unsigned int		stat_n_write_ops;
#ifdef DEEP_STATS
	unsigned int		stat_n_read_ops;
#endif
};


/**
 * To record a link between a tag value and an struct dh_data item, we also need
 * to record its bucket in the same data structure so it is covered by the same
 * lock (referencing the bucket as item->bucket is racey).
 */
struct dh_item_value_entry {
	struct dh_data			*item;
	struct dh_items_bucket		*bucket;
};

/*
 * HOW THE TAGS ARE STORED
 * -----------------------
 *
 * The tag buckets are indexed by the tag key. Tag buckets realize the granularity
 * of locks.
 *
 * Each tag bucket has a flat array of struct tag_key keys, which holds the
 * actual keys. It is expected that each tag bucket (i.e. key hash collisions) will
 * contain a small number of invididual keys (on the order of 8) for efficiency,
 * i.e. that the keypace is small.
 *
 * Each of the entries in keys contains a root of the tag_value_tree named
 * tree_head. This tree contains the values part of the tags.
 *
 * Each entry in the tag_value_tree contains a flat vitems array (which has
 * items_size elements) containing pointers to the actual data items (or NULL).
 */


/**
 * A "tag value" structure, with pointers to items containing this specific
 * key-value pair.
 */
struct tag_value {
	RB_ENTRY(tag_value)		linkage;
	signed int			value;
	unsigned int			items_size;
	struct dh_item_value_entry	*vitems;
};


static inline int
sh_tag_value_cmp(struct tag_value *a, struct tag_value *b)
{
	if (a->value == b->value)
		return 0;
	else
		return a->value < b->value ? -1 : 1;
}


/**
 * The root of a tree containing tag values for this specific key. Key must
 * be non-zero.
 */
struct tag_key {
	tag_key_t			key;
	struct tag_value_tree		tree_head;
};

RB_PROTOTYPE(tag_value_tree, tag_value, linkage, sh_tag_value_cmp);


/**
 * Each tag bucket contains an array of tag_key structures, each of which
 * contains a tree of tag_value structures.
 */
struct dh_tags_bucket {
	unsigned int			index;
	unsigned int			keys_size;
	struct tag_key			*keys;
	unsigned int			stat_n_write_ops;
#ifdef DEEP_STATS
	unsigned int			stat_n_read_ops;
#endif
};


/*
 * Locking notes:
 * Locking in the direction: item bucket lock held -> single tag locks acquired
 * and released can go in blocking mode without deadlocks. All other directions
 * and forms of locking, including acquirings of multiple locks of the same
 * level, must go through the retry (trylock) protocol to avoid deadlocks.
 */


static inline int
sh_item_cmp(struct dh_data *a, struct dh_data *b)
{
	int res;
/*
	assert(a != NULL);
	assert(b != NULL);
*/
	if (a->name_size != b->name_size)
		res = (int)(a->name_size) - (int)(b->name_size);
	else
		res = memcmp(a->name, b->name, a->name_size);
	return res;
}

struct dh_item_tree;

RB_PROTOTYPE(dh_item_tree, dh_data, linkage, __unused);
RB_PROTOTYPE(tag_value_tree, tag_value, linkage, __unused);


class MDCacheException : public std::exception {
private:
	char *msg;
public:
	MDCacheException(char *msg) : exception() {
		this->msg = msg;
	}
};


/**
 * DataStore is the principal data storage class.
 *
 * The locking dance described:
 *
 *	items_locks and tags_locks are the lock pools for items and tags; locks
 *	are for all intents and purposes identified with their numeric IDs, which
 *	are also indexes to items and tags arrays they protect.
 *
 *	Since they are abstracted as integers, they are passed around as such to
 *	other backends in the mesh (if any) by the BackendMesh via the lock
 *	notifiers.
 *
 *	The assumption is (XXX: is it still valid?) that all single lock operations
 *	must succeed (e.g. single item / tag lock operations) since they are blocking,
 *	but all multiple-lock operations must implement the trylock-and-rollback protocol
 *	to avoid deadlocks.
 *
 *	This will probably need to be revised when dealing with partition tolerance.
 */
class DataStore {
private:
	struct dh_items_bucket	items[N_ITEM_BUCKETS];
	LockPool<N_ITEM_BUCKETS> items_locks;
	struct dh_tags_bucket	tags[N_TAG_BUCKETS];
	LockPool<N_TAG_BUCKETS>	tags_locks;
	
	unsigned stat_reap_dead_count;
	unsigned stat_reap_memory_count;

	void remove_item_and_free(struct dh_data *d);
	void data_cleanup(struct dh_data *d);
	void tags_cleanup(struct dh_data *d);
	bool tag_link(struct dh_data *d, struct mc_tag *t);
	bool tag_unlink(struct dh_data *d, struct mc_tag *t);
	int get_items_buckets_by_tag_key_values(struct tag_key *tk, const tag_value_t *tvalues, const unsigned values_count,
		unsigned **buckets, unsigned *n_buckets);
	bool items_buckets_rdlock(unsigned *buckets, unsigned n_buckets);
	bool items_buckets_wrlock(unsigned *buckets, unsigned n_buckets);
	void items_buckets_unlock(unsigned *buckets, unsigned n_buckets);
	bool tags_buckets_wrlock(unsigned *buckets, unsigned n_buckets);
	void tags_buckets_unlock(unsigned *buckets, unsigned n_buckets);
	bool data_item_safe_lock_op(struct dh_data *d, enum LOCK_OP lop, bool doitem, bool dotags);
	bool multidata_append(struct mc_resp_multidata **md, struct mc_data_entry *item);
	bool multistring_append(struct mc_resp_multistring **ms, const char *str, const unsigned short int str_size);
	int put_nolock(struct dh_items_bucket *ib, struct dh_data **d);
	struct mc_data_entry* get_data_nolock(char *name, unsigned int name_size);
	int del_data_name_nolock(const char *name, const unsigned int name_size);
	int del_nolock(struct dh_items_bucket *ib, struct dh_data *d);
	int prewarm_cache_dump_bucket(FILE *f, unsigned bi);
	bool prewarm_cache_read_put_record(FILE *f, bool do_lock);
	struct dh_data *internal_add_record_nolock(const char *name, unsigned name_size, const char *data, unsigned data_size, const struct mc_tag *tags, unsigned n_tags);
	bool item_is_alive(struct dh_data *d);
public:
	unsigned int 		n_item_buckets;
	unsigned int		n_tag_buckets;

	static void fill_dh_data(struct dh_data *entry, struct mc_data_entry *inpkt);

	DataStore() throw(MDCacheException);
	~DataStore();

	void set_lock_notifiers(LockNotifier *ilock_notifier, LockNotifier *tlock_notifier);
	void unset_lock_notifiers();

	enum LOCK_RESULT _internalLockOp(enum LOCK_OP lop, unsigned *nlist, unsigned list_size, enum LOCK_TYPE);

	bool prewarm_cache_dump(char *filename, bool atomic);
	bool prewarm_cache_read(char *filename, bool atomic);

	/** Put a cooked dh_data structure into the cache */
	int put(struct dh_data *d);

	/** Get a cooked dh_data structure by name */
	struct dh_data* get(char *name, unsigned int name_len);

	/** Get a mc_data_entry structure by name */
	struct mc_data_entry* get_data(char *name, unsigned int name_size);

	/** Get a mc_data_entry structure by name but leave the tree locked for reading (to support zero-copy) */
	struct mc_data_entry* get_data_and_lock(char *name, unsigned int name_size, unsigned int *lock_cookie);

	/** Unlock the tree locked by dh_get_data_and_lock */
	void got_data_unlock(unsigned int lock_cookie);

	/** Retrieve multiple data items matching given key and any of the given values */
	struct mc_resp_multidata* get_multidata_by_tag_values(const tag_key_t tkey, tag_value_t *values, const unsigned int values_count);

	/**
	 * Retrieve all the record keys (key names) stored in the cache. Depending
	 * on the flags this may happen with all buckets locked or one by one.
	 */
	struct mc_resp_multistring *get_all_keys(const unsigned flags);

	/** Delete a data item identified by name */
	int del_data_name(const char *name, const unsigned int name_size);

	/** Delete data items identified by key and a list of values */
	int del_data_tag_values(const tag_key_t tag_key, const unsigned int num_values, const tag_value_t *values);

	/** One of the simple atomic ops. */
	struct mc_resp_atomic* atomic_op(const int op, const char *name, const unsigned int name_size, const int64_t arg1, const int64_t arg2);

	/** Multiple-put: puts multiple records into the cache */
	int mput(struct dh_data **entries, unsigned n_records, unsigned flags);

	/** Multiple-get: gets multiple records from the cache by record keys */
	struct mc_resp_multidata* mget(struct mc_key_name **keys, unsigned n_names, unsigned flags);

	/** Multi-del: deletes multile records from the cache by record keys */
	int mdel(struct mc_key_name **keys, unsigned n_names, unsigned flags);
	
	/** Tstack pop: returnes & removes a record from a virtual stack specified by the tag key & value */
	struct mc_data_entry* tstack_pop(tag_key_t tkey, tag_value_t tval);
	
	/** Perform book-keeping; returns the number of records reaped */
	unsigned reap_dead_items();
	unsigned reap_free_memory(uint64_t memory_limit);
};

#endif
